/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __E_DBUS_IDENTITY_BROKER_H__
#define __E_DBUS_IDENTITY_BROKER_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for com.microsoft.identity.Broker1 */

#define E_DBUS_TYPE_IDENTITY_BROKER1 (e_dbus_identity_broker1_get_type ())
#define E_DBUS_IDENTITY_BROKER1(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_IDENTITY_BROKER1, EDBusIdentityBroker1))
#define E_DBUS_IS_IDENTITY_BROKER1(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_IDENTITY_BROKER1))
#define E_DBUS_IDENTITY_BROKER1_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), E_DBUS_TYPE_IDENTITY_BROKER1, EDBusIdentityBroker1Iface))

struct _EDBusIdentityBroker1;
typedef struct _EDBusIdentityBroker1 EDBusIdentityBroker1;
typedef struct _EDBusIdentityBroker1Iface EDBusIdentityBroker1Iface;

struct _EDBusIdentityBroker1Iface
{
  GTypeInterface parent_iface;

  gboolean (*handle_acquire_prt_sso_cookie) (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2);

  gboolean (*handle_acquire_token_interactively) (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2);

  gboolean (*handle_acquire_token_silently) (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2);

  gboolean (*handle_cancel_interactive_flow) (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2);

  gboolean (*handle_generate_signed_http_request) (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2);

  gboolean (*handle_get_accounts) (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2);

  gboolean (*handle_get_linux_broker_version) (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2);

  gboolean (*handle_remove_account) (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2);

};

GType e_dbus_identity_broker1_get_type (void);

GDBusInterfaceInfo *e_dbus_identity_broker1_interface_info (void);
guint e_dbus_identity_broker1_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void e_dbus_identity_broker1_complete_acquire_token_interactively (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *unnamed_arg3);

void e_dbus_identity_broker1_complete_acquire_token_silently (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *unnamed_arg3);

void e_dbus_identity_broker1_complete_get_accounts (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *unnamed_arg3);

void e_dbus_identity_broker1_complete_remove_account (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *unnamed_arg3);

void e_dbus_identity_broker1_complete_acquire_prt_sso_cookie (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *unnamed_arg3);

void e_dbus_identity_broker1_complete_generate_signed_http_request (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *unnamed_arg3);

void e_dbus_identity_broker1_complete_cancel_interactive_flow (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *unnamed_arg3);

void e_dbus_identity_broker1_complete_get_linux_broker_version (
    EDBusIdentityBroker1 *object,
    GDBusMethodInvocation *invocation,
    const gchar *unnamed_arg3);



/* D-Bus method calls: */
void e_dbus_identity_broker1_call_acquire_token_interactively (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_identity_broker1_call_acquire_token_interactively_finish (
    EDBusIdentityBroker1 *proxy,
    gchar ** out_unnamed_arg3,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_identity_broker1_call_acquire_token_interactively_sync (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_unnamed_arg3,
    GCancellable *cancellable,
    GError **error);

void e_dbus_identity_broker1_call_acquire_token_silently (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_identity_broker1_call_acquire_token_silently_finish (
    EDBusIdentityBroker1 *proxy,
    gchar ** out_unnamed_arg3,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_identity_broker1_call_acquire_token_silently_sync (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_unnamed_arg3,
    GCancellable *cancellable,
    GError **error);

void e_dbus_identity_broker1_call_get_accounts (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_identity_broker1_call_get_accounts_finish (
    EDBusIdentityBroker1 *proxy,
    gchar ** out_unnamed_arg3,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_identity_broker1_call_get_accounts_sync (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_unnamed_arg3,
    GCancellable *cancellable,
    GError **error);

void e_dbus_identity_broker1_call_remove_account (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_identity_broker1_call_remove_account_finish (
    EDBusIdentityBroker1 *proxy,
    gchar ** out_unnamed_arg3,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_identity_broker1_call_remove_account_sync (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_unnamed_arg3,
    GCancellable *cancellable,
    GError **error);

void e_dbus_identity_broker1_call_acquire_prt_sso_cookie (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_identity_broker1_call_acquire_prt_sso_cookie_finish (
    EDBusIdentityBroker1 *proxy,
    gchar ** out_unnamed_arg3,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_identity_broker1_call_acquire_prt_sso_cookie_sync (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_unnamed_arg3,
    GCancellable *cancellable,
    GError **error);

void e_dbus_identity_broker1_call_generate_signed_http_request (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_identity_broker1_call_generate_signed_http_request_finish (
    EDBusIdentityBroker1 *proxy,
    gchar ** out_unnamed_arg3,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_identity_broker1_call_generate_signed_http_request_sync (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_unnamed_arg3,
    GCancellable *cancellable,
    GError **error);

void e_dbus_identity_broker1_call_cancel_interactive_flow (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_identity_broker1_call_cancel_interactive_flow_finish (
    EDBusIdentityBroker1 *proxy,
    gchar ** out_unnamed_arg3,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_identity_broker1_call_cancel_interactive_flow_sync (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_unnamed_arg3,
    GCancellable *cancellable,
    GError **error);

void e_dbus_identity_broker1_call_get_linux_broker_version (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_identity_broker1_call_get_linux_broker_version_finish (
    EDBusIdentityBroker1 *proxy,
    gchar ** out_unnamed_arg3,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_identity_broker1_call_get_linux_broker_version_sync (
    EDBusIdentityBroker1 *proxy,
    const gchar *arg_unnamed_arg0,
    const gchar *arg_unnamed_arg1,
    const gchar *arg_unnamed_arg2,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_unnamed_arg3,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define E_DBUS_TYPE_IDENTITY_BROKER1_PROXY (e_dbus_identity_broker1_proxy_get_type ())
#define E_DBUS_IDENTITY_BROKER1_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_IDENTITY_BROKER1_PROXY, EDBusIdentityBroker1Proxy))
#define E_DBUS_IDENTITY_BROKER1_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_IDENTITY_BROKER1_PROXY, EDBusIdentityBroker1ProxyClass))
#define E_DBUS_IDENTITY_BROKER1_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_IDENTITY_BROKER1_PROXY, EDBusIdentityBroker1ProxyClass))
#define E_DBUS_IS_IDENTITY_BROKER1_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_IDENTITY_BROKER1_PROXY))
#define E_DBUS_IS_IDENTITY_BROKER1_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_IDENTITY_BROKER1_PROXY))

typedef struct _EDBusIdentityBroker1Proxy EDBusIdentityBroker1Proxy;
typedef struct _EDBusIdentityBroker1ProxyClass EDBusIdentityBroker1ProxyClass;
typedef struct _EDBusIdentityBroker1ProxyPrivate EDBusIdentityBroker1ProxyPrivate;

struct _EDBusIdentityBroker1Proxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  EDBusIdentityBroker1ProxyPrivate *priv;
};

struct _EDBusIdentityBroker1ProxyClass
{
  GDBusProxyClass parent_class;
};

GType e_dbus_identity_broker1_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusIdentityBroker1Proxy, g_object_unref)
#endif

void e_dbus_identity_broker1_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusIdentityBroker1 *e_dbus_identity_broker1_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusIdentityBroker1 *e_dbus_identity_broker1_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void e_dbus_identity_broker1_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusIdentityBroker1 *e_dbus_identity_broker1_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusIdentityBroker1 *e_dbus_identity_broker1_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define E_DBUS_TYPE_IDENTITY_BROKER1_SKELETON (e_dbus_identity_broker1_skeleton_get_type ())
#define E_DBUS_IDENTITY_BROKER1_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_IDENTITY_BROKER1_SKELETON, EDBusIdentityBroker1Skeleton))
#define E_DBUS_IDENTITY_BROKER1_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_IDENTITY_BROKER1_SKELETON, EDBusIdentityBroker1SkeletonClass))
#define E_DBUS_IDENTITY_BROKER1_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_IDENTITY_BROKER1_SKELETON, EDBusIdentityBroker1SkeletonClass))
#define E_DBUS_IS_IDENTITY_BROKER1_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_IDENTITY_BROKER1_SKELETON))
#define E_DBUS_IS_IDENTITY_BROKER1_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_IDENTITY_BROKER1_SKELETON))

typedef struct _EDBusIdentityBroker1Skeleton EDBusIdentityBroker1Skeleton;
typedef struct _EDBusIdentityBroker1SkeletonClass EDBusIdentityBroker1SkeletonClass;
typedef struct _EDBusIdentityBroker1SkeletonPrivate EDBusIdentityBroker1SkeletonPrivate;

struct _EDBusIdentityBroker1Skeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  EDBusIdentityBroker1SkeletonPrivate *priv;
};

struct _EDBusIdentityBroker1SkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType e_dbus_identity_broker1_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusIdentityBroker1Skeleton, g_object_unref)
#endif

EDBusIdentityBroker1 *e_dbus_identity_broker1_skeleton_new (void);


G_END_DECLS

#endif /* __E_DBUS_IDENTITY_BROKER_H__ */
