/*
 * Copyright (C) 2006 OpenedHand Ltd
 * Copyright (C) 2009 Intel Corporation
 *
 * This library is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors: Ross Burton <ross@linux.intel.com>
 */

#ifndef EDS_DISABLE_DEPRECATED

/* Do not generate bindings. */
#ifndef __GI_SCANNER__

#ifndef E_BOOK_VIEW_PRIVATE_H
#define E_BOOK_VIEW_PRIVATE_H

#include "e-book.h"
#include "e-book-client.h"
#include "e-book-client-view.h"
#include "e-book-view.h"

G_BEGIN_DECLS

EBookView *	_e_book_view_new		(EBook *book,
						 EBookClientView *client_view);

G_END_DECLS

#endif /* E_BOOK_VIEW_PRIVATE_H */

#endif /* __GI_SCANNER__ */

#endif /* EDS_DISABLE_DEPRECATED */

