
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "e-book-contacts-enumtypes.h"
/* enumerations from "e-book-contacts-enums.h" */
#include "e-book-contacts-enums.h"
GType
e_book_client_view_flags_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_BOOK_CLIENT_VIEW_FLAGS_NONE,
			  "E_BOOK_CLIENT_VIEW_FLAGS_NONE",
			  "none" },
			{ E_BOOK_CLIENT_VIEW_FLAGS_NOTIFY_INITIAL,
			  "E_BOOK_CLIENT_VIEW_FLAGS_NOTIFY_INITIAL",
			  "notify-initial" },
			{ E_BOOK_CLIENT_VIEW_FLAGS_MANUAL_QUERY,
			  "E_BOOK_CLIENT_VIEW_FLAGS_MANUAL_QUERY",
			  "manual-query" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("EBookClientViewFlags"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_book_index_type_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_BOOK_INDEX_PREFIX,
			  "E_BOOK_INDEX_PREFIX",
			  "prefix" },
			{ E_BOOK_INDEX_SUFFIX,
			  "E_BOOK_INDEX_SUFFIX",
			  "suffix" },
			{ E_BOOK_INDEX_PHONE,
			  "E_BOOK_INDEX_PHONE",
			  "phone" },
			{ E_BOOK_INDEX_SORT_KEY,
			  "E_BOOK_INDEX_SORT_KEY",
			  "sort-key" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("EBookIndexType"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_book_cursor_sort_type_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_BOOK_CURSOR_SORT_ASCENDING,
			  "E_BOOK_CURSOR_SORT_ASCENDING",
			  "ascending" },
			{ E_BOOK_CURSOR_SORT_DESCENDING,
			  "E_BOOK_CURSOR_SORT_DESCENDING",
			  "descending" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("EBookCursorSortType"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_book_cursor_origin_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_BOOK_CURSOR_ORIGIN_CURRENT,
			  "E_BOOK_CURSOR_ORIGIN_CURRENT",
			  "current" },
			{ E_BOOK_CURSOR_ORIGIN_BEGIN,
			  "E_BOOK_CURSOR_ORIGIN_BEGIN",
			  "begin" },
			{ E_BOOK_CURSOR_ORIGIN_END,
			  "E_BOOK_CURSOR_ORIGIN_END",
			  "end" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("EBookCursorOrigin"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_book_cursor_step_flags_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_BOOK_CURSOR_STEP_MOVE,
			  "E_BOOK_CURSOR_STEP_MOVE",
			  "move" },
			{ E_BOOK_CURSOR_STEP_FETCH,
			  "E_BOOK_CURSOR_STEP_FETCH",
			  "fetch" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("EBookCursorStepFlags"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_book_operation_flags_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_BOOK_OPERATION_FLAG_NONE,
			  "E_BOOK_OPERATION_FLAG_NONE",
			  "none" },
			{ E_BOOK_OPERATION_FLAG_CONFLICT_FAIL,
			  "E_BOOK_OPERATION_FLAG_CONFLICT_FAIL",
			  "conflict-fail" },
			{ E_BOOK_OPERATION_FLAG_CONFLICT_USE_NEWER,
			  "E_BOOK_OPERATION_FLAG_CONFLICT_USE_NEWER",
			  "conflict-use-newer" },
			{ E_BOOK_OPERATION_FLAG_CONFLICT_KEEP_SERVER,
			  "E_BOOK_OPERATION_FLAG_CONFLICT_KEEP_SERVER",
			  "conflict-keep-server" },
			{ E_BOOK_OPERATION_FLAG_CONFLICT_KEEP_LOCAL,
			  "E_BOOK_OPERATION_FLAG_CONFLICT_KEEP_LOCAL",
			  "conflict-keep-local" },
			{ E_BOOK_OPERATION_FLAG_CONFLICT_WRITE_COPY,
			  "E_BOOK_OPERATION_FLAG_CONFLICT_WRITE_COPY",
			  "conflict-write-copy" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("EBookOperationFlags"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_vcard_version_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_VCARD_VERSION_UNKNOWN,
			  "E_VCARD_VERSION_UNKNOWN",
			  "unknown" },
			{ E_VCARD_VERSION_21,
			  "E_VCARD_VERSION_21",
			  "21" },
			{ E_VCARD_VERSION_30,
			  "E_VCARD_VERSION_30",
			  "30" },
			{ E_VCARD_VERSION_40,
			  "E_VCARD_VERSION_40",
			  "40" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("EVCardVersion"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_contact_date_time_flags_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_CONTACT_DATE_TIME_FLAG_NONE,
			  "E_CONTACT_DATE_TIME_FLAG_NONE",
			  "none" },
			{ E_CONTACT_DATE_TIME_FLAG_DATE_TIME,
			  "E_CONTACT_DATE_TIME_FLAG_DATE_TIME",
			  "date-time" },
			{ E_CONTACT_DATE_TIME_FLAG_TIME,
			  "E_CONTACT_DATE_TIME_FLAG_TIME",
			  "time" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("EContactDateTimeFlags"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_vcard_foreach_flags_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_VCARD_FOREACH_FLAG_NONE,
			  "E_VCARD_FOREACH_FLAG_NONE",
			  "none" },
			{ E_VCARD_FOREACH_FLAG_WILL_MODIFY,
			  "E_VCARD_FOREACH_FLAG_WILL_MODIFY",
			  "will-modify" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("EVCardForeachFlags"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}

/* Generated data ends here */

