/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __E_DBUS_OAUTH2_RESPONSE_H__
#define __E_DBUS_OAUTH2_RESPONSE_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.evolution.dataserver.OAuth2Response */

#define E_DBUS_TYPE_OAUTH2_RESPONSE (e_dbus_oauth2_response_get_type ())
#define E_DBUS_OAUTH2_RESPONSE(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_OAUTH2_RESPONSE, EDBusOAuth2Response))
#define E_DBUS_IS_OAUTH2_RESPONSE(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_OAUTH2_RESPONSE))
#define E_DBUS_OAUTH2_RESPONSE_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), E_DBUS_TYPE_OAUTH2_RESPONSE, EDBusOAuth2ResponseIface))

struct _EDBusOAuth2Response;
typedef struct _EDBusOAuth2Response EDBusOAuth2Response;
typedef struct _EDBusOAuth2ResponseIface EDBusOAuth2ResponseIface;

struct _EDBusOAuth2ResponseIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_response_uri) (
    EDBusOAuth2Response *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_uri);

};

GType e_dbus_oauth2_response_get_type (void);

GDBusInterfaceInfo *e_dbus_oauth2_response_interface_info (void);
guint e_dbus_oauth2_response_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void e_dbus_oauth2_response_complete_response_uri (
    EDBusOAuth2Response *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void e_dbus_oauth2_response_call_response_uri (
    EDBusOAuth2Response *proxy,
    const gchar *arg_uri,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_oauth2_response_call_response_uri_finish (
    EDBusOAuth2Response *proxy,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_oauth2_response_call_response_uri_sync (
    EDBusOAuth2Response *proxy,
    const gchar *arg_uri,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define E_DBUS_TYPE_OAUTH2_RESPONSE_PROXY (e_dbus_oauth2_response_proxy_get_type ())
#define E_DBUS_OAUTH2_RESPONSE_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_OAUTH2_RESPONSE_PROXY, EDBusOAuth2ResponseProxy))
#define E_DBUS_OAUTH2_RESPONSE_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_OAUTH2_RESPONSE_PROXY, EDBusOAuth2ResponseProxyClass))
#define E_DBUS_OAUTH2_RESPONSE_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_OAUTH2_RESPONSE_PROXY, EDBusOAuth2ResponseProxyClass))
#define E_DBUS_IS_OAUTH2_RESPONSE_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_OAUTH2_RESPONSE_PROXY))
#define E_DBUS_IS_OAUTH2_RESPONSE_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_OAUTH2_RESPONSE_PROXY))

typedef struct _EDBusOAuth2ResponseProxy EDBusOAuth2ResponseProxy;
typedef struct _EDBusOAuth2ResponseProxyClass EDBusOAuth2ResponseProxyClass;
typedef struct _EDBusOAuth2ResponseProxyPrivate EDBusOAuth2ResponseProxyPrivate;

struct _EDBusOAuth2ResponseProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  EDBusOAuth2ResponseProxyPrivate *priv;
};

struct _EDBusOAuth2ResponseProxyClass
{
  GDBusProxyClass parent_class;
};

GType e_dbus_oauth2_response_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusOAuth2ResponseProxy, g_object_unref)
#endif

void e_dbus_oauth2_response_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusOAuth2Response *e_dbus_oauth2_response_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusOAuth2Response *e_dbus_oauth2_response_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void e_dbus_oauth2_response_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusOAuth2Response *e_dbus_oauth2_response_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusOAuth2Response *e_dbus_oauth2_response_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define E_DBUS_TYPE_OAUTH2_RESPONSE_SKELETON (e_dbus_oauth2_response_skeleton_get_type ())
#define E_DBUS_OAUTH2_RESPONSE_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_OAUTH2_RESPONSE_SKELETON, EDBusOAuth2ResponseSkeleton))
#define E_DBUS_OAUTH2_RESPONSE_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_OAUTH2_RESPONSE_SKELETON, EDBusOAuth2ResponseSkeletonClass))
#define E_DBUS_OAUTH2_RESPONSE_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_OAUTH2_RESPONSE_SKELETON, EDBusOAuth2ResponseSkeletonClass))
#define E_DBUS_IS_OAUTH2_RESPONSE_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_OAUTH2_RESPONSE_SKELETON))
#define E_DBUS_IS_OAUTH2_RESPONSE_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_OAUTH2_RESPONSE_SKELETON))

typedef struct _EDBusOAuth2ResponseSkeleton EDBusOAuth2ResponseSkeleton;
typedef struct _EDBusOAuth2ResponseSkeletonClass EDBusOAuth2ResponseSkeletonClass;
typedef struct _EDBusOAuth2ResponseSkeletonPrivate EDBusOAuth2ResponseSkeletonPrivate;

struct _EDBusOAuth2ResponseSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  EDBusOAuth2ResponseSkeletonPrivate *priv;
};

struct _EDBusOAuth2ResponseSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType e_dbus_oauth2_response_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusOAuth2ResponseSkeleton, g_object_unref)
#endif

EDBusOAuth2Response *e_dbus_oauth2_response_skeleton_new (void);


G_END_DECLS

#endif /* __E_DBUS_OAUTH2_RESPONSE_H__ */
