
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "e-backend-enumtypes.h"
/* enumerations from "e-backend-enums.h" */
#include "e-backend-enums.h"
GType
e_authentication_session_result_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_AUTHENTICATION_SESSION_ERROR,
			  "E_AUTHENTICATION_SESSION_ERROR",
			  "error" },
			{ E_AUTHENTICATION_SESSION_SUCCESS,
			  "E_AUTHENTICATION_SESSION_SUCCESS",
			  "success" },
			{ E_AUTHENTICATION_SESSION_DISMISSED,
			  "E_AUTHENTICATION_SESSION_DISMISSED",
			  "dismissed" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("EAuthenticationSessionResult"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_dbus_server_exit_code_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_DBUS_SERVER_EXIT_NONE,
			  "E_DBUS_SERVER_EXIT_NONE",
			  "none" },
			{ E_DBUS_SERVER_EXIT_NORMAL,
			  "E_DBUS_SERVER_EXIT_NORMAL",
			  "normal" },
			{ E_DBUS_SERVER_EXIT_RELOAD,
			  "E_DBUS_SERVER_EXIT_RELOAD",
			  "reload" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("EDBusServerExitCode"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_source_permission_flags_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_SOURCE_PERMISSION_NONE,
			  "E_SOURCE_PERMISSION_NONE",
			  "none" },
			{ E_SOURCE_PERMISSION_WRITABLE,
			  "E_SOURCE_PERMISSION_WRITABLE",
			  "writable" },
			{ E_SOURCE_PERMISSION_REMOVABLE,
			  "E_SOURCE_PERMISSION_REMOVABLE",
			  "removable" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("ESourcePermissionFlags"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_offline_state_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_OFFLINE_STATE_UNKNOWN,
			  "E_OFFLINE_STATE_UNKNOWN",
			  "unknown" },
			{ E_OFFLINE_STATE_SYNCED,
			  "E_OFFLINE_STATE_SYNCED",
			  "synced" },
			{ E_OFFLINE_STATE_LOCALLY_CREATED,
			  "E_OFFLINE_STATE_LOCALLY_CREATED",
			  "locally-created" },
			{ E_OFFLINE_STATE_LOCALLY_MODIFIED,
			  "E_OFFLINE_STATE_LOCALLY_MODIFIED",
			  "locally-modified" },
			{ E_OFFLINE_STATE_LOCALLY_DELETED,
			  "E_OFFLINE_STATE_LOCALLY_DELETED",
			  "locally-deleted" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("EOfflineState"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_collection_backend_parts_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_COLLECTION_BACKEND_PART_NONE,
			  "E_COLLECTION_BACKEND_PART_NONE",
			  "none" },
			{ E_COLLECTION_BACKEND_PART_CALENDAR,
			  "E_COLLECTION_BACKEND_PART_CALENDAR",
			  "calendar" },
			{ E_COLLECTION_BACKEND_PART_CONTACTS,
			  "E_COLLECTION_BACKEND_PART_CONTACTS",
			  "contacts" },
			{ E_COLLECTION_BACKEND_PART_MAIL,
			  "E_COLLECTION_BACKEND_PART_MAIL",
			  "mail" },
			{ E_COLLECTION_BACKEND_PART_ANY,
			  "E_COLLECTION_BACKEND_PART_ANY",
			  "any" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("ECollectionBackendParts"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}

/* Generated data ends here */

