/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __E_DBUS_CALENDAR_FACTORY_H__
#define __E_DBUS_CALENDAR_FACTORY_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.evolution.dataserver.CalendarFactory */

#define E_DBUS_TYPE_CALENDAR_FACTORY (e_dbus_calendar_factory_get_type ())
#define E_DBUS_CALENDAR_FACTORY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_CALENDAR_FACTORY, EDBusCalendarFactory))
#define E_DBUS_IS_CALENDAR_FACTORY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_CALENDAR_FACTORY))
#define E_DBUS_CALENDAR_FACTORY_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), E_DBUS_TYPE_CALENDAR_FACTORY, EDBusCalendarFactoryIface))

struct _EDBusCalendarFactory;
typedef struct _EDBusCalendarFactory EDBusCalendarFactory;
typedef struct _EDBusCalendarFactoryIface EDBusCalendarFactoryIface;

struct _EDBusCalendarFactoryIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_open_calendar) (
    EDBusCalendarFactory *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_source_uid);

  gboolean (*handle_open_memo_list) (
    EDBusCalendarFactory *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_source_uid);

  gboolean (*handle_open_task_list) (
    EDBusCalendarFactory *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_source_uid);

};

GType e_dbus_calendar_factory_get_type (void);

GDBusInterfaceInfo *e_dbus_calendar_factory_interface_info (void);
guint e_dbus_calendar_factory_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void e_dbus_calendar_factory_complete_open_calendar (
    EDBusCalendarFactory *object,
    GDBusMethodInvocation *invocation,
    const gchar *object_path,
    const gchar *bus_name);

void e_dbus_calendar_factory_complete_open_task_list (
    EDBusCalendarFactory *object,
    GDBusMethodInvocation *invocation,
    const gchar *object_path,
    const gchar *bus_name);

void e_dbus_calendar_factory_complete_open_memo_list (
    EDBusCalendarFactory *object,
    GDBusMethodInvocation *invocation,
    const gchar *object_path,
    const gchar *bus_name);



/* D-Bus method calls: */
void e_dbus_calendar_factory_call_open_calendar (
    EDBusCalendarFactory *proxy,
    const gchar *arg_source_uid,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_calendar_factory_call_open_calendar_finish (
    EDBusCalendarFactory *proxy,
    gchar ** out_object_path,
    gchar ** out_bus_name,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_calendar_factory_call_open_calendar_sync (
    EDBusCalendarFactory *proxy,
    const gchar *arg_source_uid,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_object_path,
    gchar ** out_bus_name,
    GCancellable *cancellable,
    GError **error);

void e_dbus_calendar_factory_call_open_task_list (
    EDBusCalendarFactory *proxy,
    const gchar *arg_source_uid,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_calendar_factory_call_open_task_list_finish (
    EDBusCalendarFactory *proxy,
    gchar ** out_object_path,
    gchar ** out_bus_name,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_calendar_factory_call_open_task_list_sync (
    EDBusCalendarFactory *proxy,
    const gchar *arg_source_uid,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_object_path,
    gchar ** out_bus_name,
    GCancellable *cancellable,
    GError **error);

void e_dbus_calendar_factory_call_open_memo_list (
    EDBusCalendarFactory *proxy,
    const gchar *arg_source_uid,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_calendar_factory_call_open_memo_list_finish (
    EDBusCalendarFactory *proxy,
    gchar ** out_object_path,
    gchar ** out_bus_name,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_calendar_factory_call_open_memo_list_sync (
    EDBusCalendarFactory *proxy,
    const gchar *arg_source_uid,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_object_path,
    gchar ** out_bus_name,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define E_DBUS_TYPE_CALENDAR_FACTORY_PROXY (e_dbus_calendar_factory_proxy_get_type ())
#define E_DBUS_CALENDAR_FACTORY_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_CALENDAR_FACTORY_PROXY, EDBusCalendarFactoryProxy))
#define E_DBUS_CALENDAR_FACTORY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_CALENDAR_FACTORY_PROXY, EDBusCalendarFactoryProxyClass))
#define E_DBUS_CALENDAR_FACTORY_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_CALENDAR_FACTORY_PROXY, EDBusCalendarFactoryProxyClass))
#define E_DBUS_IS_CALENDAR_FACTORY_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_CALENDAR_FACTORY_PROXY))
#define E_DBUS_IS_CALENDAR_FACTORY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_CALENDAR_FACTORY_PROXY))

typedef struct _EDBusCalendarFactoryProxy EDBusCalendarFactoryProxy;
typedef struct _EDBusCalendarFactoryProxyClass EDBusCalendarFactoryProxyClass;
typedef struct _EDBusCalendarFactoryProxyPrivate EDBusCalendarFactoryProxyPrivate;

struct _EDBusCalendarFactoryProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  EDBusCalendarFactoryProxyPrivate *priv;
};

struct _EDBusCalendarFactoryProxyClass
{
  GDBusProxyClass parent_class;
};

GType e_dbus_calendar_factory_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusCalendarFactoryProxy, g_object_unref)
#endif

void e_dbus_calendar_factory_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusCalendarFactory *e_dbus_calendar_factory_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusCalendarFactory *e_dbus_calendar_factory_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void e_dbus_calendar_factory_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusCalendarFactory *e_dbus_calendar_factory_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusCalendarFactory *e_dbus_calendar_factory_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define E_DBUS_TYPE_CALENDAR_FACTORY_SKELETON (e_dbus_calendar_factory_skeleton_get_type ())
#define E_DBUS_CALENDAR_FACTORY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_CALENDAR_FACTORY_SKELETON, EDBusCalendarFactorySkeleton))
#define E_DBUS_CALENDAR_FACTORY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_CALENDAR_FACTORY_SKELETON, EDBusCalendarFactorySkeletonClass))
#define E_DBUS_CALENDAR_FACTORY_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_CALENDAR_FACTORY_SKELETON, EDBusCalendarFactorySkeletonClass))
#define E_DBUS_IS_CALENDAR_FACTORY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_CALENDAR_FACTORY_SKELETON))
#define E_DBUS_IS_CALENDAR_FACTORY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_CALENDAR_FACTORY_SKELETON))

typedef struct _EDBusCalendarFactorySkeleton EDBusCalendarFactorySkeleton;
typedef struct _EDBusCalendarFactorySkeletonClass EDBusCalendarFactorySkeletonClass;
typedef struct _EDBusCalendarFactorySkeletonPrivate EDBusCalendarFactorySkeletonPrivate;

struct _EDBusCalendarFactorySkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  EDBusCalendarFactorySkeletonPrivate *priv;
};

struct _EDBusCalendarFactorySkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType e_dbus_calendar_factory_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusCalendarFactorySkeleton, g_object_unref)
#endif

EDBusCalendarFactory *e_dbus_calendar_factory_skeleton_new (void);


G_END_DECLS

#endif /* __E_DBUS_CALENDAR_FACTORY_H__ */
