
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef CAMEL_ENUMTYPES_H
#define CAMEL_ENUMTYPES_H

#include <glib-object.h>

G_BEGIN_DECLS

/* Enumerations from "camel-enums.h" */
#define CAMEL_TYPE_AUTHENTICATION_RESULT	(camel_authentication_result_get_type())
GType camel_authentication_result_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_FOLDER_FLAGS	(camel_folder_flags_get_type())
GType camel_folder_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_FOLDER_INFO_FLAGS	(camel_folder_info_flags_get_type())
GType camel_folder_info_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_MIME_FILTER_TO_HTML_FLAGS	(camel_mime_filter_to_html_flags_get_type())
GType camel_mime_filter_to_html_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_STORE_INFO_FLAGS	(camel_store_info_flags_get_type())
GType camel_store_info_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_FETCH_HEADERS_TYPE	(camel_fetch_headers_type_get_type())
GType camel_fetch_headers_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_JUNK_STATUS	(camel_junk_status_get_type())
GType camel_junk_status_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_MIME_FILTER_BASIC_TYPE	(camel_mime_filter_basic_type_get_type())
GType camel_mime_filter_basic_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_MIME_FILTER_CRLF_DIRECTION	(camel_mime_filter_crlf_direction_get_type())
GType camel_mime_filter_crlf_direction_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_MIME_FILTER_CRLF_MODE	(camel_mime_filter_crlf_mode_get_type())
GType camel_mime_filter_crlf_mode_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_MIME_FILTER_GZIP_MODE	(camel_mime_filter_gzip_mode_get_type())
GType camel_mime_filter_gzip_mode_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_MIME_FILTER_YENC_DIRECTION	(camel_mime_filter_yenc_direction_get_type())
GType camel_mime_filter_yenc_direction_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_NETWORK_SECURITY_METHOD	(camel_network_security_method_get_type())
GType camel_network_security_method_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_PROVIDER_CONF_TYPE	(camel_provider_conf_type_get_type())
GType camel_provider_conf_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_PROVIDER_FLAGS	(camel_provider_flags_get_type())
GType camel_provider_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_PROVIDER_TYPE	(camel_provider_type_get_type())
GType camel_provider_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_SASL_ANON_TRACE_TYPE	(camel_sasl_anon_trace_type_get_type())
GType camel_sasl_anon_trace_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_RECIPIENT_CERTIFICATE_FLAGS	(camel_recipient_certificate_flags_get_type())
GType camel_recipient_certificate_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_SERVICE_CONNECTION_STATUS	(camel_service_connection_status_get_type())
GType camel_service_connection_status_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_SESSION_ALERT_TYPE	(camel_session_alert_type_get_type())
GType camel_session_alert_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_SORT_TYPE	(camel_sort_type_get_type())
GType camel_sort_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_STORE_FLAGS	(camel_store_flags_get_type())
GType camel_store_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_STORE_GET_FOLDER_INFO_FLAGS	(camel_store_get_folder_info_flags_get_type())
GType camel_store_get_folder_info_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_STORE_PERMISSION_FLAGS	(camel_store_permission_flags_get_type())
GType camel_store_permission_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_TRANSFER_ENCODING	(camel_transfer_encoding_get_type())
GType camel_transfer_encoding_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_THREE_STATE	(camel_three_state_get_type())
GType camel_three_state_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_COMPARE_TYPE	(camel_compare_type_get_type())
GType camel_compare_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_TIME_UNIT	(camel_time_unit_get_type())
GType camel_time_unit_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_GPG_TRUST	(camel_gpg_trust_get_type())
GType camel_gpg_trust_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_STORE_DB_COUNT_KIND	(camel_store_db_count_kind_get_type())
GType camel_store_db_count_kind_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_DB_SQLIZE_FLAGS	(camel_db_sqlize_flags_get_type())
GType camel_db_sqlize_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_FOLDER_THREAD_FLAGS	(camel_folder_thread_flags_get_type())
GType camel_folder_thread_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_MATCH_THREADS_KIND	(camel_match_threads_kind_get_type())
GType camel_match_threads_kind_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_VEE_FOLDER_OP_FLAGS	(camel_vee_folder_op_flags_get_type())
GType camel_vee_folder_op_flags_get_type	(void) G_GNUC_CONST;

/* Enumerations from "camel-folder.h" */
#define CAMEL_TYPE_FOLDER_ERROR	(camel_folder_error_get_type())
GType camel_folder_error_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_FOLDER_PARAM_FLAGS	(camel_folder_param_flags_get_type())
GType camel_folder_param_flags_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_FETCH_TYPE	(camel_fetch_type_get_type())
GType camel_fetch_type_get_type	(void) G_GNUC_CONST;

/* Enumerations from "camel-message-info.h" */
#define CAMEL_TYPE_MESSAGE_FLAGS	(camel_message_flags_get_type())
GType camel_message_flags_get_type	(void) G_GNUC_CONST;

/* Enumerations from "camel-mime-filter-canon.h" */
#define CAMEL_TYPE_MIME_FILTER_CANON_FLAGS	(camel_mime_filter_canon_flags_get_type())
GType camel_mime_filter_canon_flags_get_type	(void) G_GNUC_CONST;

/* Enumerations from "camel-mime-filter-enriched.h" */
#define CAMEL_TYPE_MIME_FILTER_ENRICHED_FLAGS	(camel_mime_filter_enriched_flags_get_type())
GType camel_mime_filter_enriched_flags_get_type	(void) G_GNUC_CONST;

/* Enumerations from "camel-service.h" */
#define CAMEL_TYPE_SERVICE_ERROR	(camel_service_error_get_type())
GType camel_service_error_get_type	(void) G_GNUC_CONST;

/* Enumerations from "camel-sexp.h" */
#define CAMEL_TYPE_SEXP_RESULT_TYPE	(camel_sexp_result_type_get_type())
GType camel_sexp_result_type_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_SEXP_TERM_TYPE	(camel_sexp_term_type_get_type())
GType camel_sexp_term_type_get_type	(void) G_GNUC_CONST;

/* Enumerations from "camel-store.h" */
#define CAMEL_TYPE_STORE_ERROR	(camel_store_error_get_type())
GType camel_store_error_get_type	(void) G_GNUC_CONST;
#define CAMEL_TYPE_STORE_GET_FOLDER_FLAGS	(camel_store_get_folder_flags_get_type())
GType camel_store_get_folder_flags_get_type	(void) G_GNUC_CONST;

/* Enumerations from "camel-url.h" */
#define CAMEL_TYPE_URL_FLAGS	(camel_url_flags_get_type())
GType camel_url_flags_get_type	(void) G_GNUC_CONST;

G_END_DECLS

#endif /* CAMEL_ENUMTYPES_H */

/* Generated data ends here */

