/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __E_DBUS_CALENDAR_VIEW_H__
#define __E_DBUS_CALENDAR_VIEW_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.evolution.dataserver.CalendarView */

#define E_DBUS_TYPE_CALENDAR_VIEW (e_dbus_calendar_view_get_type ())
#define E_DBUS_CALENDAR_VIEW(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_CALENDAR_VIEW, EDBusCalendarView))
#define E_DBUS_IS_CALENDAR_VIEW(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_CALENDAR_VIEW))
#define E_DBUS_CALENDAR_VIEW_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), E_DBUS_TYPE_CALENDAR_VIEW, EDBusCalendarViewIface))

struct _EDBusCalendarView;
typedef struct _EDBusCalendarView EDBusCalendarView;
typedef struct _EDBusCalendarViewIface EDBusCalendarViewIface;

struct _EDBusCalendarViewIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_dispose) (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_set_fields_of_interest) (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_only_fields);

  gboolean (*handle_set_flags) (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation,
    guint arg_flags);

  gboolean (*handle_start) (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_stop) (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation);

  void (*complete) (
    EDBusCalendarView *object,
    const gchar *arg_error_name,
    const gchar *arg_error_message);

  void (*objects_added) (
    EDBusCalendarView *object,
    const gchar *const *arg_objects);

  void (*objects_modified) (
    EDBusCalendarView *object,
    const gchar *const *arg_objects);

  void (*objects_removed) (
    EDBusCalendarView *object,
    const gchar *const *arg_uids);

  void (*progress) (
    EDBusCalendarView *object,
    guint arg_percent,
    const gchar *arg_message);

};

GType e_dbus_calendar_view_get_type (void);

GDBusInterfaceInfo *e_dbus_calendar_view_interface_info (void);
guint e_dbus_calendar_view_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void e_dbus_calendar_view_complete_start (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation);

void e_dbus_calendar_view_complete_stop (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation);

void e_dbus_calendar_view_complete_set_flags (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation);

void e_dbus_calendar_view_complete_set_fields_of_interest (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation);

void e_dbus_calendar_view_complete_dispose (
    EDBusCalendarView *object,
    GDBusMethodInvocation *invocation);



/* D-Bus signal emissions functions: */
void e_dbus_calendar_view_emit_objects_added (
    EDBusCalendarView *object,
    const gchar *const *arg_objects);

void e_dbus_calendar_view_emit_objects_modified (
    EDBusCalendarView *object,
    const gchar *const *arg_objects);

void e_dbus_calendar_view_emit_objects_removed (
    EDBusCalendarView *object,
    const gchar *const *arg_uids);

void e_dbus_calendar_view_emit_progress (
    EDBusCalendarView *object,
    guint arg_percent,
    const gchar *arg_message);

void e_dbus_calendar_view_emit_complete (
    EDBusCalendarView *object,
    const gchar *arg_error_name,
    const gchar *arg_error_message);



/* D-Bus method calls: */
void e_dbus_calendar_view_call_start (
    EDBusCalendarView *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_calendar_view_call_start_finish (
    EDBusCalendarView *proxy,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_calendar_view_call_start_sync (
    EDBusCalendarView *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void e_dbus_calendar_view_call_stop (
    EDBusCalendarView *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_calendar_view_call_stop_finish (
    EDBusCalendarView *proxy,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_calendar_view_call_stop_sync (
    EDBusCalendarView *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void e_dbus_calendar_view_call_set_flags (
    EDBusCalendarView *proxy,
    guint arg_flags,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_calendar_view_call_set_flags_finish (
    EDBusCalendarView *proxy,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_calendar_view_call_set_flags_sync (
    EDBusCalendarView *proxy,
    guint arg_flags,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void e_dbus_calendar_view_call_set_fields_of_interest (
    EDBusCalendarView *proxy,
    const gchar *const *arg_only_fields,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_calendar_view_call_set_fields_of_interest_finish (
    EDBusCalendarView *proxy,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_calendar_view_call_set_fields_of_interest_sync (
    EDBusCalendarView *proxy,
    const gchar *const *arg_only_fields,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void e_dbus_calendar_view_call_dispose (
    EDBusCalendarView *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_calendar_view_call_dispose_finish (
    EDBusCalendarView *proxy,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_calendar_view_call_dispose_sync (
    EDBusCalendarView *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define E_DBUS_TYPE_CALENDAR_VIEW_PROXY (e_dbus_calendar_view_proxy_get_type ())
#define E_DBUS_CALENDAR_VIEW_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_CALENDAR_VIEW_PROXY, EDBusCalendarViewProxy))
#define E_DBUS_CALENDAR_VIEW_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_CALENDAR_VIEW_PROXY, EDBusCalendarViewProxyClass))
#define E_DBUS_CALENDAR_VIEW_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_CALENDAR_VIEW_PROXY, EDBusCalendarViewProxyClass))
#define E_DBUS_IS_CALENDAR_VIEW_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_CALENDAR_VIEW_PROXY))
#define E_DBUS_IS_CALENDAR_VIEW_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_CALENDAR_VIEW_PROXY))

typedef struct _EDBusCalendarViewProxy EDBusCalendarViewProxy;
typedef struct _EDBusCalendarViewProxyClass EDBusCalendarViewProxyClass;
typedef struct _EDBusCalendarViewProxyPrivate EDBusCalendarViewProxyPrivate;

struct _EDBusCalendarViewProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  EDBusCalendarViewProxyPrivate *priv;
};

struct _EDBusCalendarViewProxyClass
{
  GDBusProxyClass parent_class;
};

GType e_dbus_calendar_view_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusCalendarViewProxy, g_object_unref)
#endif

void e_dbus_calendar_view_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusCalendarView *e_dbus_calendar_view_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusCalendarView *e_dbus_calendar_view_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void e_dbus_calendar_view_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusCalendarView *e_dbus_calendar_view_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusCalendarView *e_dbus_calendar_view_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define E_DBUS_TYPE_CALENDAR_VIEW_SKELETON (e_dbus_calendar_view_skeleton_get_type ())
#define E_DBUS_CALENDAR_VIEW_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_CALENDAR_VIEW_SKELETON, EDBusCalendarViewSkeleton))
#define E_DBUS_CALENDAR_VIEW_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_CALENDAR_VIEW_SKELETON, EDBusCalendarViewSkeletonClass))
#define E_DBUS_CALENDAR_VIEW_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_CALENDAR_VIEW_SKELETON, EDBusCalendarViewSkeletonClass))
#define E_DBUS_IS_CALENDAR_VIEW_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_CALENDAR_VIEW_SKELETON))
#define E_DBUS_IS_CALENDAR_VIEW_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_CALENDAR_VIEW_SKELETON))

typedef struct _EDBusCalendarViewSkeleton EDBusCalendarViewSkeleton;
typedef struct _EDBusCalendarViewSkeletonClass EDBusCalendarViewSkeletonClass;
typedef struct _EDBusCalendarViewSkeletonPrivate EDBusCalendarViewSkeletonPrivate;

struct _EDBusCalendarViewSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  EDBusCalendarViewSkeletonPrivate *priv;
};

struct _EDBusCalendarViewSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType e_dbus_calendar_view_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusCalendarViewSkeleton, g_object_unref)
#endif

EDBusCalendarView *e_dbus_calendar_view_skeleton_new (void);


G_END_DECLS

#endif /* __E_DBUS_CALENDAR_VIEW_H__ */
