
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "e-cal-enumtypes.h"
/* enumerations from "e-cal-enums.h" */
#include "e-cal-enums.h"
GType
e_cal_client_source_type_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_CLIENT_SOURCE_TYPE_EVENTS,
			  "E_CAL_CLIENT_SOURCE_TYPE_EVENTS",
			  "events" },
			{ E_CAL_CLIENT_SOURCE_TYPE_TASKS,
			  "E_CAL_CLIENT_SOURCE_TYPE_TASKS",
			  "tasks" },
			{ E_CAL_CLIENT_SOURCE_TYPE_MEMOS,
			  "E_CAL_CLIENT_SOURCE_TYPE_MEMOS",
			  "memos" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalClientSourceType"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_obj_mod_type_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_CAL_OBJ_MOD_THIS,
			  "E_CAL_OBJ_MOD_THIS",
			  "this" },
			{ E_CAL_OBJ_MOD_THIS_AND_PRIOR,
			  "E_CAL_OBJ_MOD_THIS_AND_PRIOR",
			  "this-and-prior" },
			{ E_CAL_OBJ_MOD_THIS_AND_FUTURE,
			  "E_CAL_OBJ_MOD_THIS_AND_FUTURE",
			  "this-and-future" },
			{ E_CAL_OBJ_MOD_ALL,
			  "E_CAL_OBJ_MOD_ALL",
			  "all" },
			{ E_CAL_OBJ_MOD_ONLY_THIS,
			  "E_CAL_OBJ_MOD_ONLY_THIS",
			  "only-this" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("ECalObjModType"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_component_vtype_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_COMPONENT_NO_TYPE,
			  "E_CAL_COMPONENT_NO_TYPE",
			  "no-type" },
			{ E_CAL_COMPONENT_EVENT,
			  "E_CAL_COMPONENT_EVENT",
			  "event" },
			{ E_CAL_COMPONENT_TODO,
			  "E_CAL_COMPONENT_TODO",
			  "todo" },
			{ E_CAL_COMPONENT_JOURNAL,
			  "E_CAL_COMPONENT_JOURNAL",
			  "journal" },
			{ E_CAL_COMPONENT_FREEBUSY,
			  "E_CAL_COMPONENT_FREEBUSY",
			  "freebusy" },
			{ E_CAL_COMPONENT_TIMEZONE,
			  "E_CAL_COMPONENT_TIMEZONE",
			  "timezone" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalComponentVType"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_component_classification_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_COMPONENT_CLASS_NONE,
			  "E_CAL_COMPONENT_CLASS_NONE",
			  "none" },
			{ E_CAL_COMPONENT_CLASS_PUBLIC,
			  "E_CAL_COMPONENT_CLASS_PUBLIC",
			  "public" },
			{ E_CAL_COMPONENT_CLASS_PRIVATE,
			  "E_CAL_COMPONENT_CLASS_PRIVATE",
			  "private" },
			{ E_CAL_COMPONENT_CLASS_CONFIDENTIAL,
			  "E_CAL_COMPONENT_CLASS_CONFIDENTIAL",
			  "confidential" },
			{ E_CAL_COMPONENT_CLASS_UNKNOWN,
			  "E_CAL_COMPONENT_CLASS_UNKNOWN",
			  "unknown" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalComponentClassification"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_component_period_kind_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_COMPONENT_PERIOD_DATETIME,
			  "E_CAL_COMPONENT_PERIOD_DATETIME",
			  "datetime" },
			{ E_CAL_COMPONENT_PERIOD_DURATION,
			  "E_CAL_COMPONENT_PERIOD_DURATION",
			  "duration" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalComponentPeriodKind"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_component_range_kind_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_COMPONENT_RANGE_SINGLE,
			  "E_CAL_COMPONENT_RANGE_SINGLE",
			  "single" },
			{ E_CAL_COMPONENT_RANGE_THISPRIOR,
			  "E_CAL_COMPONENT_RANGE_THISPRIOR",
			  "thisprior" },
			{ E_CAL_COMPONENT_RANGE_THISFUTURE,
			  "E_CAL_COMPONENT_RANGE_THISFUTURE",
			  "thisfuture" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalComponentRangeKind"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_component_transparency_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_COMPONENT_TRANSP_NONE,
			  "E_CAL_COMPONENT_TRANSP_NONE",
			  "none" },
			{ E_CAL_COMPONENT_TRANSP_TRANSPARENT,
			  "E_CAL_COMPONENT_TRANSP_TRANSPARENT",
			  "transparent" },
			{ E_CAL_COMPONENT_TRANSP_OPAQUE,
			  "E_CAL_COMPONENT_TRANSP_OPAQUE",
			  "opaque" },
			{ E_CAL_COMPONENT_TRANSP_UNKNOWN,
			  "E_CAL_COMPONENT_TRANSP_UNKNOWN",
			  "unknown" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalComponentTransparency"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_component_alarm_action_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_COMPONENT_ALARM_NONE,
			  "E_CAL_COMPONENT_ALARM_NONE",
			  "none" },
			{ E_CAL_COMPONENT_ALARM_AUDIO,
			  "E_CAL_COMPONENT_ALARM_AUDIO",
			  "audio" },
			{ E_CAL_COMPONENT_ALARM_DISPLAY,
			  "E_CAL_COMPONENT_ALARM_DISPLAY",
			  "display" },
			{ E_CAL_COMPONENT_ALARM_EMAIL,
			  "E_CAL_COMPONENT_ALARM_EMAIL",
			  "email" },
			{ E_CAL_COMPONENT_ALARM_PROCEDURE,
			  "E_CAL_COMPONENT_ALARM_PROCEDURE",
			  "procedure" },
			{ E_CAL_COMPONENT_ALARM_UNKNOWN,
			  "E_CAL_COMPONENT_ALARM_UNKNOWN",
			  "unknown" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalComponentAlarmAction"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_component_alarm_trigger_kind_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_COMPONENT_ALARM_TRIGGER_NONE,
			  "E_CAL_COMPONENT_ALARM_TRIGGER_NONE",
			  "none" },
			{ E_CAL_COMPONENT_ALARM_TRIGGER_RELATIVE_START,
			  "E_CAL_COMPONENT_ALARM_TRIGGER_RELATIVE_START",
			  "relative-start" },
			{ E_CAL_COMPONENT_ALARM_TRIGGER_RELATIVE_END,
			  "E_CAL_COMPONENT_ALARM_TRIGGER_RELATIVE_END",
			  "relative-end" },
			{ E_CAL_COMPONENT_ALARM_TRIGGER_ABSOLUTE,
			  "E_CAL_COMPONENT_ALARM_TRIGGER_ABSOLUTE",
			  "absolute" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalComponentAlarmTriggerKind"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_recur_describe_recurrence_flags_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_CAL_RECUR_DESCRIBE_RECURRENCE_FLAG_NONE,
			  "E_CAL_RECUR_DESCRIBE_RECURRENCE_FLAG_NONE",
			  "none" },
			{ E_CAL_RECUR_DESCRIBE_RECURRENCE_FLAG_PREFIXED,
			  "E_CAL_RECUR_DESCRIBE_RECURRENCE_FLAG_PREFIXED",
			  "prefixed" },
			{ E_CAL_RECUR_DESCRIBE_RECURRENCE_FLAG_FALLBACK,
			  "E_CAL_RECUR_DESCRIBE_RECURRENCE_FLAG_FALLBACK",
			  "fallback" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("ECalRecurDescribeRecurrenceFlags"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_operation_flags_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GFlagsValue values[] = {
			{ E_CAL_OPERATION_FLAG_NONE,
			  "E_CAL_OPERATION_FLAG_NONE",
			  "none" },
			{ E_CAL_OPERATION_FLAG_CONFLICT_FAIL,
			  "E_CAL_OPERATION_FLAG_CONFLICT_FAIL",
			  "conflict-fail" },
			{ E_CAL_OPERATION_FLAG_CONFLICT_USE_NEWER,
			  "E_CAL_OPERATION_FLAG_CONFLICT_USE_NEWER",
			  "conflict-use-newer" },
			{ E_CAL_OPERATION_FLAG_CONFLICT_KEEP_SERVER,
			  "E_CAL_OPERATION_FLAG_CONFLICT_KEEP_SERVER",
			  "conflict-keep-server" },
			{ E_CAL_OPERATION_FLAG_CONFLICT_KEEP_LOCAL,
			  "E_CAL_OPERATION_FLAG_CONFLICT_KEEP_LOCAL",
			  "conflict-keep-local" },
			{ E_CAL_OPERATION_FLAG_CONFLICT_WRITE_COPY,
			  "E_CAL_OPERATION_FLAG_CONFLICT_WRITE_COPY",
			  "conflict-write-copy" },
			{ E_CAL_OPERATION_FLAG_DISABLE_ITIP_MESSAGE,
			  "E_CAL_OPERATION_FLAG_DISABLE_ITIP_MESSAGE",
			  "disable-itip-message" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_flags_register_static (
			g_intern_static_string ("ECalOperationFlags"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}
GType
e_cal_interval_units_get_type (void)
{
	static gsize the_type__volatile = 0;

	if (g_once_init_enter (&the_type__volatile)) {
		static const GEnumValue values[] = {
			{ E_CAL_INTERVAL_UNIT_NONE,
			  "E_CAL_INTERVAL_UNIT_NONE",
			  "none" },
			{ E_CAL_INTERVAL_UNIT_MINUTES,
			  "E_CAL_INTERVAL_UNIT_MINUTES",
			  "minutes" },
			{ E_CAL_INTERVAL_UNIT_HOURS,
			  "E_CAL_INTERVAL_UNIT_HOURS",
			  "hours" },
			{ E_CAL_INTERVAL_UNIT_DAYS,
			  "E_CAL_INTERVAL_UNIT_DAYS",
			  "days" },
			{ 0, NULL, NULL }
		};
		GType the_type = g_enum_register_static (
			g_intern_static_string ("ECalIntervalUnits"),
			values);
		g_once_init_leave (&the_type__volatile, the_type);
	}
	return the_type__volatile;
}

/* Generated data ends here */

