/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __E_DBUS_USER_PROMPTER_H__
#define __E_DBUS_USER_PROMPTER_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.evolution.dataserver.UserPrompter */

#define E_DBUS_TYPE_USER_PROMPTER (e_dbus_user_prompter_get_type ())
#define E_DBUS_USER_PROMPTER(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_USER_PROMPTER, EDBusUserPrompter))
#define E_DBUS_IS_USER_PROMPTER(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_USER_PROMPTER))
#define E_DBUS_USER_PROMPTER_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), E_DBUS_TYPE_USER_PROMPTER, EDBusUserPrompterIface))

struct _EDBusUserPrompter;
typedef struct _EDBusUserPrompter EDBusUserPrompter;
typedef struct _EDBusUserPrompterIface EDBusUserPrompterIface;

struct _EDBusUserPrompterIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_extension_prompt) (
    EDBusUserPrompter *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_dialog_name,
    const gchar *const *arg_parameters);

  gboolean (*handle_prompt) (
    EDBusUserPrompter *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_type,
    const gchar *arg_title,
    const gchar *arg_primary_text,
    const gchar *arg_secondary_text,
    gboolean arg_use_markup,
    const gchar *const *arg_button_captions);

  void (*extension_response) (
    EDBusUserPrompter *object,
    gint arg_id,
    gint arg_response,
    const gchar *const *arg_values);

  void (*response) (
    EDBusUserPrompter *object,
    gint arg_id,
    gint arg_response_button);

};

GType e_dbus_user_prompter_get_type (void);

GDBusInterfaceInfo *e_dbus_user_prompter_interface_info (void);
guint e_dbus_user_prompter_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void e_dbus_user_prompter_complete_prompt (
    EDBusUserPrompter *object,
    GDBusMethodInvocation *invocation,
    gint id);

void e_dbus_user_prompter_complete_extension_prompt (
    EDBusUserPrompter *object,
    GDBusMethodInvocation *invocation,
    gint id);



/* D-Bus signal emissions functions: */
void e_dbus_user_prompter_emit_response (
    EDBusUserPrompter *object,
    gint arg_id,
    gint arg_response_button);

void e_dbus_user_prompter_emit_extension_response (
    EDBusUserPrompter *object,
    gint arg_id,
    gint arg_response,
    const gchar *const *arg_values);



/* D-Bus method calls: */
void e_dbus_user_prompter_call_prompt (
    EDBusUserPrompter *proxy,
    const gchar *arg_type,
    const gchar *arg_title,
    const gchar *arg_primary_text,
    const gchar *arg_secondary_text,
    gboolean arg_use_markup,
    const gchar *const *arg_button_captions,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_user_prompter_call_prompt_finish (
    EDBusUserPrompter *proxy,
    gint* out_id,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_user_prompter_call_prompt_sync (
    EDBusUserPrompter *proxy,
    const gchar *arg_type,
    const gchar *arg_title,
    const gchar *arg_primary_text,
    const gchar *arg_secondary_text,
    gboolean arg_use_markup,
    const gchar *const *arg_button_captions,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gint* out_id,
    GCancellable *cancellable,
    GError **error);

void e_dbus_user_prompter_call_extension_prompt (
    EDBusUserPrompter *proxy,
    const gchar *arg_dialog_name,
    const gchar *const *arg_parameters,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean e_dbus_user_prompter_call_extension_prompt_finish (
    EDBusUserPrompter *proxy,
    gint* out_id,
    GAsyncResult *res,
    GError **error);

gboolean e_dbus_user_prompter_call_extension_prompt_sync (
    EDBusUserPrompter *proxy,
    const gchar *arg_dialog_name,
    const gchar *const *arg_parameters,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gint* out_id,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define E_DBUS_TYPE_USER_PROMPTER_PROXY (e_dbus_user_prompter_proxy_get_type ())
#define E_DBUS_USER_PROMPTER_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_USER_PROMPTER_PROXY, EDBusUserPrompterProxy))
#define E_DBUS_USER_PROMPTER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_USER_PROMPTER_PROXY, EDBusUserPrompterProxyClass))
#define E_DBUS_USER_PROMPTER_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_USER_PROMPTER_PROXY, EDBusUserPrompterProxyClass))
#define E_DBUS_IS_USER_PROMPTER_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_USER_PROMPTER_PROXY))
#define E_DBUS_IS_USER_PROMPTER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_USER_PROMPTER_PROXY))

typedef struct _EDBusUserPrompterProxy EDBusUserPrompterProxy;
typedef struct _EDBusUserPrompterProxyClass EDBusUserPrompterProxyClass;
typedef struct _EDBusUserPrompterProxyPrivate EDBusUserPrompterProxyPrivate;

struct _EDBusUserPrompterProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  EDBusUserPrompterProxyPrivate *priv;
};

struct _EDBusUserPrompterProxyClass
{
  GDBusProxyClass parent_class;
};

GType e_dbus_user_prompter_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusUserPrompterProxy, g_object_unref)
#endif

void e_dbus_user_prompter_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusUserPrompter *e_dbus_user_prompter_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusUserPrompter *e_dbus_user_prompter_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void e_dbus_user_prompter_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
EDBusUserPrompter *e_dbus_user_prompter_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
EDBusUserPrompter *e_dbus_user_prompter_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define E_DBUS_TYPE_USER_PROMPTER_SKELETON (e_dbus_user_prompter_skeleton_get_type ())
#define E_DBUS_USER_PROMPTER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), E_DBUS_TYPE_USER_PROMPTER_SKELETON, EDBusUserPrompterSkeleton))
#define E_DBUS_USER_PROMPTER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), E_DBUS_TYPE_USER_PROMPTER_SKELETON, EDBusUserPrompterSkeletonClass))
#define E_DBUS_USER_PROMPTER_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), E_DBUS_TYPE_USER_PROMPTER_SKELETON, EDBusUserPrompterSkeletonClass))
#define E_DBUS_IS_USER_PROMPTER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), E_DBUS_TYPE_USER_PROMPTER_SKELETON))
#define E_DBUS_IS_USER_PROMPTER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), E_DBUS_TYPE_USER_PROMPTER_SKELETON))

typedef struct _EDBusUserPrompterSkeleton EDBusUserPrompterSkeleton;
typedef struct _EDBusUserPrompterSkeletonClass EDBusUserPrompterSkeletonClass;
typedef struct _EDBusUserPrompterSkeletonPrivate EDBusUserPrompterSkeletonPrivate;

struct _EDBusUserPrompterSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  EDBusUserPrompterSkeletonPrivate *priv;
};

struct _EDBusUserPrompterSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType e_dbus_user_prompter_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (EDBusUserPrompterSkeleton, g_object_unref)
#endif

EDBusUserPrompter *e_dbus_user_prompter_skeleton_new (void);


G_END_DECLS

#endif /* __E_DBUS_USER_PROMPTER_H__ */
